from pydantic import BaseModel
from typing import Literal, Union


class W24Coordinate(BaseModel):
    x: int
    y: int


class W24GeometryEllipse(BaseModel):
    """
    Representation of an Ellipse with center coordinates,
    major and minor radius, and an initial angle of rotation.

    Attributes
    ----------
    center (W24Coordinate): The center coordinates of the ellipse (x, y)
    major_radius (float): Major Radius of the Ellipse
    minor_radius (float): Minor Radius of the Ellipse
    angle (float): Angle of Rotation for the Ellipse (default is 0.0)
    """

    geometry_type: Literal["ELLIPSE"] = "ELLIPSE"

    center: W24Coordinate
    major_radius: float
    minor_radius: float
    angle: float = 0.0


class W24GeometryLine(BaseModel):
    """
    Representation of a Line with start and end coordinates.

    Attributes
    ----------
    start (W24Coordinate): Start coordinates of the Line (x, y)
    end (W24Coordinate): End coordinates of the Line (x, y)
    """

    geometry_type: Literal["LINE"] = "LINE"

    start: W24Coordinate
    end: W24Coordinate


W24GeometryFeature = Union[W24GeometryEllipse, W24GeometryLine]


class W24PositionedFeature(BaseModel):
    """
    Representation of the Ellipse placement on the different thumbnails.

    Attributes:
    ----------
    sheet (W24Ellipse): Position of the Geometrz Feature on he Sheet Thumbnail
    """

    sheet: W24GeometryFeature



from enum import Enum


class W24FraunhoferLine(str, Enum):
    """
    In optics, the "d" and "e" lines refer to specific spectral lines
    associated with the emission spectrum of hydrogen. They are used
    as standard reference points for calibrating optical instruments
    and for characterizing optical materials like glass and lenses.

    The "d-line" typically refers to the Fraunhofer line, which
    is the major spectral line in the yellow region of the
    spectrum emitted by sodium (Na). It is actually a doublet,
    with wavelengths approximately at 589.0 nm (D2) and 589.6 nm (D1).
    The d-line is often used in the testing of optical components
    due to its narrow bandwidth and its position in the spectrum
    close to the maximum sensitivity of the human eye.

    The "e-line" is also a spectral line associated with mercury (Hg)
    emission. It has a wavelength of about 546.1 nm and falls within
    the green part of the spectrum. The e-line is also used in the
    characterization of optical systems.
    """

    D_LINE = "D_LINE"
    E_LINE = "E_LINE"



""" HTTPS-part of the Werk24 client
"""

import uuid
from werk24.exceptions import SSLCertificateError
import json
import urllib.parse
from pydantic import UUID4
from werk24.models.ask import W24AskUnion
from typing import List
from types import TracebackType
from typing import Dict, Optional, Type
from io import BufferedReader
import aiohttp
from pydantic import HttpUrl
from typing import Union
from werk24.auth_client import AuthClient
from werk24.exceptions import (
    BadRequestException,
    RequestTooLargeException,
    ResourceNotFoundException,
    ServerException,
    UnauthorizedException,
    UnsupportedMediaType,
)
import ssl
import certifi
from werk24.models.techread import W24TechreadWithCallbackPayload
from werk24.models.helpdesk import W24HelpdeskTask
from werk24.models.techread import W24PresignedPost
from werk24._version import __version__

EXCEPTION_CLASSES = {
    range(200, 300): None,
    range(400, 401): BadRequestException,
    range(401, 403): UnauthorizedException,
    range(404, 405): ResourceNotFoundException,
    range(413, 414): RequestTooLargeException,
    range(415, 416): UnsupportedMediaType,
    range(300, 400): ServerException,
    range(500, 600): ServerException,
    range(416, 500): ServerException,
}


class TechreadClientHttps:
    """Translation map from the server response
    to the W24TechreadArchitectureStatus enum
    """

    def __init__(self, techread_version: str, support_base_url: str):
        """
        Initialize a new session with the https server.

        Arguments:
            techread_server_https {str} -- Domain of the Techread https server
            techread_version {str} -- Techread Version
            support_base_url {str} -- Base URL for support requests
        """

        self._techread_version = techread_version
        self._auth_client: Optional[AuthClient] = None
        self.support_base_url = support_base_url

    def _make_session(self, timeout_seconds=30) -> aiohttp.TCPConnector:
        """Make the connector for the session.

        Returns:
        --------
        aiohttp.TCPConnector: Connector for the session
        """
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = aiohttp.TCPConnector(ssl=ssl_context)
        timeout = aiohttp.ClientTimeout(
            total=None,
            sock_connect=timeout_seconds,
            sock_read=timeout_seconds,
        )
        return aiohttp.ClientSession(
            timeout=timeout,
            connector=connector,
        )

    async def __aenter__(self) -> "TechreadClientHttps":
        """
        Create a new HTTP session that is being used for the whole connection.
        Be sure to keep the session alive.

        Raises:
        ------
        RuntimeError: Exception raised when the developer tries to start the
            session without a token.

        Returns:
        -------
        TechreadClientHttps: Instance of the class itself with active session.
        """
        if self._auth_client is None:
            raise RuntimeError("No AuthClient was registered")

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Close the session.

        Args:
        ----
        exc_type (Optional[Type[BaseException]]): The type of exception that
            caused the context manager to be exited.
        exc_value (Optional[BaseException]): The instance of the exception that
            caused the exit.
        traceback (Optional[TracebackType]): A traceback from the exception.

        Returns:
        --------
        None
        """
        pass

    def register_auth_client(self, auth_client: AuthClient) -> None:
        """Register the reference to the authentication service

        Arguments:
            auth_client {AuthClient} -- Reference to Authentication
                client
        """
        self._auth_client = auth_client

    async def upload_associated_file(
        self,
        presigned_post: W24PresignedPost,
        content: Optional[Union[bytes, BufferedReader]],
    ) -> None:
        """
        Uploads an associated file to the API.

        This can either be a technical drawing or a 3D model.
        NOTE: The complete message size must not be larger than 10 MB.

        Args:
        ----
        presigned_post (W24PresignedPost): Presigned post object for
            file upload.
        content (Optional[bytes]): Content of the file as bytes.

        Raises:
        -------
        Various exceptions based on the issues with API, authentication
            or the requested file.

        Returns:
        --------
        None
        """
        # ignore if payload is empty
        if content is None:
            return

        # generate the form data by merging the presigned
        # fields with the file
        form = aiohttp.FormData()
        for key, value in presigned_post.fields_.items():
            form.add_field(key, value)
        form.add_field("file", content)

        # create a new fresh session that does not
        # carry the authentication token
        presigned_post_str = str(presigned_post.url)
        try:
            async with self._make_session() as session:
                async with session.post(presigned_post_str, data=form) as response:
                    self._raise_for_status(presigned_post_str, response.status)

        # Raise SSLCertificateError if the certificate is not trusted
        except aiohttp.ClientConnectorCertificateError as exception:
            raise SSLCertificateError() from exception

    async def download_payload(self, payload_url: HttpUrl) -> bytes:
        """Return the payload from the server

        Args:
        ----
        payload_url {HttpUrl} -- Url of the payload

        Raises:
        ------
        RuntimeError: Hard Error that is raised when
            the function is asked to download a payload
            from an untrusted source.
            This provides some sort of protection against
            payload-injection and token-theft. When you
            see this error showing up, you should
            definitely INVESTIGATE AND LET US KNOW
            IMMEDIATELY!!!
            Call all our numbers on a Sunday morning
            at 3am if it must be. Even if its Christmas
            and Easter on the same day.

        BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!

        UnauthorizedException: Raised when the token
            or the requested file have expired

        ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.

        RequestTooLargeException: Raised when the status
            code was 413

        UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read (because its media type
            is not supported by the API).

        ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        bytes -- Payload
        """
        # send the get request to the endpoint
        try:
            async with self._make_session() as session:
                response = await session.get(str(payload_url))
                self._raise_for_status(payload_url, response.status)
                return await response.content.read()

        # reraise the exceptions
        except (
            UnauthorizedException,  # pylint: disable=try-except-raise
            RequestTooLargeException,
            ServerException,
            BadRequestException,
            ResourceNotFoundException,
        ):
            raise

    @staticmethod
    def _raise_for_status(url: str, status_code: int) -> None:
        """Raise the correct exception depending on the status code.

        Args:
        ----
        url (str): The requested URL
        status_code (int): The received response status code

        Raises:
        ------
        BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!

        UnauthorizedException: Raised when the token
            or the requested file have expired

        ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.

        RequestTooLargeException: Raised when the status
            code was 413

        UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read(because its media type
            is not supported by the API).

        ServerException: Raised for all other status codes
            that are not 2xx
        """
        for key, exception_class in EXCEPTION_CLASSES.items():
            if status_code in key:
                if exception_class is not None:
                    raise exception_class(
                        f"Request failed '{url}' with code {status_code}"
                    )
                return None

        # If the resposne code is anything other than unauthorized or 200 (OK), we trigger a ServerException.
        raise ServerException(f"Request failed '{url}' with code {status_code}")

    async def create_helpdesk_task(self, task: W24HelpdeskTask) -> W24HelpdeskTask:
        """
        Create a Helpdesk ticket.

        Args:
        ----
        task (W24HelpdeskTask): Helpdesk task to be created

        Raises:
        ------
        BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!

        UnauthorizedException: Raised when the token
            or the requested file have expired

        ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.

        RequestTooLargeException: Raised when the status
            code was 413

        UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read(because its media type
            is not supported by the API).

        ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        W24HelpdeskTask: Created helpdesk task with an updated task_id.
        """
        headers = self._make_helpdesk_headers()
        url = self._make_support_url("helpdesk/create-task")
        async with self._make_session() as session:
            response = await session.post(url, json=task.json(), headers=headers)
            self._raise_for_status(url, response.status)

        # return the updated task
        return W24HelpdeskTask.parse_raw(await response.text())

    def _make_support_url(self, path: str) -> str:
        """Make the support url for the help desk requests.

        Args:
        ----
        path (str): Path to the endpoint

        Returns:
        -------
        str: URL to the endpoint
        """
        return urllib.parse.urljoin(f"https://{self.support_base_url}", path)

    def _make_helpdesk_headers(self) -> Dict[str, str]:
        """
        Make the headers for the help desk requests.

        Simply the authorization header at this stage.

        Returns:
        -------
        Dict[str, str]: Help desk headers
        """
        return self._auth_client.get_headers()

    async def read_drawing_with_callback(
        self,
        drawing: Union[BufferedReader, bytes],
        asks: List[W24AskUnion],
        callback_url: str,
        max_pages: int = 5,
        drawing_filename: Optional[str] = None,
        callback_headers: Optional[Dict[str, str]] = None,
    ) -> UUID4:
        """
        Read a drawing with a callback.

        Args:
        ----
        drawing (Union[BufferedReader, bytes]): Drawing to be read
        asks (List[W24Ask]): List of asks
        callback_url (str): Callback URL
        max_pages (int, optional): Maximum number of pages to be read.
            Defaults to 5.
        drawing_filename (Optional[str], optional): Filename of the drawing.
            Defaults to None.
        callback_headers (Optional[Dict[str, str]], optional): Headers for the
            callback. Defaults to None.

        Raises:
        ------
        BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!

        UnauthorizedException: Raised when the token
            or the requested file have expired

        ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.

        RequestTooLargeException: Raised when the status
            code was 413

        UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read(because its media type
            is not supported by the API).

        ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        UUID4: Request ID
        """

        # Set a default drawing filename if none is provided
        drawing_filename = drawing_filename or "drawing.pdf"

        # validate the payload locally. This is not strictly necessary
        # but it is a good way to catch errors early.
        payload = W24TechreadWithCallbackPayload(
            asks=asks,
            callback_url=callback_url,
            callback_headers=callback_headers,
            max_pages=max_pages,
            client_version=__version__,
            drawing_filename=drawing_filename,
        )

        # create the form data
        data = aiohttp.FormData()
        data.add_field("drawing", drawing, filename=drawing_filename)
        for key, value in payload.model_dump(mode="json").items():
            data.add_field(key, json.dumps(value))

        # send the request
        headers = self._auth_client.get_auth_headers()
        url = self._make_support_url("techread/read-with-callback")
        async with self._make_session() as session:
            response = await session.post(url, data=data, headers=headers)
            self._raise_for_status(url, response.status)
            response_json = await response.json(content_type=None)

        try:
            return uuid.UUID(response_json["request_id"])
        except (ValueError, KeyError):
            raise BadRequestException(f"Request failed: {response_json}")



from enum import Enum
from typing import Optional

from pydantic import UUID4, BaseModel

from werk24.models.base_feature import W24BaseFeaturePosition


class W24HelpdeskImportance(str, Enum):
    """List of possible importance levels for a helpdesk ticket.
    """
    MUST_HAVE = "MUST_HAVE"
    SHOULD_HAVE = "SHOULD_HAVE"
    COULD_HAVE = "COULD_HAVE"


class W24HelpdeskTask(BaseModel):
    """
    Minimalistic HelpDesk Ticket Model.

    Attributes:
    ----------
    task_id (Optional[str]): Unique identifier of the task (if already created).
    request_id (UUID4): Unique identifier of the request.
    observed_outcome (str): Observed outcome that the user wants to report.
    expected_outcome (str): Expected outcome that the user considers correct.
    comment (str): Additional comment for explaining the issue.
    position (W24BaseFeaturePosition): Position of the issue in the document.
        When the position of the feature is known, it can be provided here.
        Leaving it empty is also accepted.
    importance (W24HelpdeskImportance): Importance of the issue.
    """
    task_id: Optional[str] = None
    request_id: UUID4
    observed_outcome: str
    expected_outcome: str
    comment: str
    position: Optional[W24BaseFeaturePosition] = None
    importance: W24HelpdeskImportance



from pydantic import BaseModel, Field
from pydantic_extra_types.color import Color
from werk24.models.alphabet import W24Alphabet


class W24Font(BaseModel):
    """Font Specification.

    Allows you to specify which font shall be used.

    """

    font_family: str = Field(
        description="Font Family",
    )
    font_style: str = Field(
        description="Font Style",
        examples=["Regular", "Medium"],
        default="Regular",
    )
    font_size: float = Field(
        description="Font Size in pt.",
        examples=[10, 15],
    )
    font_color: Color = Field(
        description="Font Color",
        default=Color((0, 0, 0)),
    )


class W24FontMap(BaseModel):
    """Object that maps alphabets to fonts.

    This allows you to support multiple alphabets
    e.g., one for Latin characters and one for
    Chinese Hanzi. Be aware that the font selection
    is performed word-by-word.

    NOTE: you always need to specify a Latin script,
    as decimals are considered to be part of the Latin
    script. (Courtesy of!? the developers of the Brahmi script).
    See: https://en.wikipedia.org/wiki/Brahmi_numerals
    """

    font_map: dict[W24Alphabet, W24Font] = Field(
        description="Dictionary that maps an alphabet to a font",
        examples={W24Alphabet.LATIN: W24Font(font_family="Work Sans", font_size=10)},
    )



from typing import Literal

from werk24.models.property.base import W24Property
from enum import Enum


class W24PropertyCleanness(W24Property):
    """Cleanness Property.
    """
    property_type: Literal["CLEANNESS"] = "CLEANNESS"


class W24PropertyCleannessTypeDin8592(str, Enum):
    CLEANED = "CLEANED"
    SCALE_REMOVED = "SCALE_REMOVED"
    RUST_REMOVED = "RUST_REMOVED"
    COATING_REMOVED = "COATING_REMOVED"
    GREASE_REMOVED = "GREASE_REMOVED"
    DUST_REMOVED = "DUST_REMOVED"
    SHOOT_REMOVED = "SHOOT_REMOVED"
    STERILIZED = "STERILIZED"
    DISINFECTED = "DISINFECTED"
    DECONTAMINATED = "DECONTAMINATED"


class W24PropertyCleannessDin8592(W24PropertyCleanness):
    """DIN 8592 Cleanness Specification.

    DIN 8592 specifies a whole range of
    cleanness targets. These cleanness
    targets are specific to the type
    of impurity that shall be removed.

    Free of Rust for example does not
    mean that the part is sterilized.

    NOTE: Future implementation will likely
    include the cleanness level as well. Be
    aware that they are measured differently
    in different industries / countries.

    Attributes:
        cleanness_type (W24PropertyCleannessTypeDin8592):
            Cleanness Type for which the cleanness
            level should be achieved.
    """
    property_subtype: Literal["DIN_8592"] = "DIN_8592"
    cleanness_type: W24PropertyCleannessTypeDin8592



from enum import Enum


class W24Alphabet(str, Enum):
    """List of Alphabets

    This lists all alphabets that the Werk24 API
    can differentiate. If you are processing
    Technical Drawings in another alphabet, please
    give us a call.

    Additional Alphabets include:
    * Armenian
    * Canadian Syllabic
    * Ethiopic
    * Georgian
    * North Indic
    * Sough Indic
    * Thaana

    """

    ARABIC = "ARABIC"
    CJK = "CJK"
    CYRILLIC = "CYRILLIC"
    GREEK = "GREEK"
    HANGUL = "HANGUL"
    HEBREW = "HEBREW"
    HIRAGANA = "HIRAGANA"
    KATAKANA = "KATAKANA"
    LATIN = "LATIN"
    THAI = "THAI"



__all__ = [
    "abbe_number",
    "bubbles_and_inclusions",
    "cleanness",
    "color",
    "corrosion_resistance",
    "fraunhofer",
    "glass_homogeneity",
    "hardness",
    "material",
    "refractive_index",
    "stress_birefringence",
    "striae",
    "surface_area",
    "weight",
]



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, validator

from werk24.models.bom_table import W24BomTable
from werk24.models.property.color import W24PropertyColor

from .base_feature import W24BaseFeatureModel
from .general_tolerances import W24GeneralTolerances
from .language import W24Language
from .material import W24Material
from .roughness import W24GeneralRoughness, W24RoughnessReference
from .unit import W24UnitSpecification
from .weight import W24Weight


class W24TitleBlockItem(W24BaseFeatureModel):
    """Per-Language caption or value

    Attributes:

        language: Language in accordance with the ISO/639-2B standards

        text: Text of the identification
    """

    language: Optional[W24Language]

    text: str


class W24CaptionValuePair(BaseModel):
    """Caption-Value pair for that were found on the Title Block.

    Attributes:

        blurb: Caption-Value pair for human consumption

        captions: List of captions in different languages.
            This will only return the languages that were detected and
            NOT translate the captions into languages that are not present
            on the drawing. This behavior might however change in the
            future.
    """

    blurb: str

    captions: List[W24TitleBlockItem]
    values: List[W24TitleBlockItem]


class W24IdentifierType(str, Enum):
    """List of Identifier Types supported by Werk24"""

    ASSEMBLY_NAME = "ASSEMBLY_NAME "
    ASSEMBLY_NUMBER = "ASSEMBLY_NUMBER"
    CAGE_CODE = "CAGE_CODE"
    CONTRACT_NUMBER = "CONTRACT_NUMBER"
    CUSTOMER_NAME = "CUSTOMER_NAME"
    CUSTOMER_NUMBER = "CUSTOMER_NUMBER"
    DOCUMENT_NUMBER = "DOCUMENT_NUMBER"
    DRAWING_NUMBER = "DRAWING_NUMBER"
    ERP_NUMBER = "ERP_NUMBER"
    IDENTIFICATION_NUMBER = "IDENTIFICATION_NUMBER"
    ITEM_NUMER = "ITEM_NUMER"
    MANUFACTURER_NAME = "MANUFACTURER_NAME"
    MANUFACTURER_NUMBER = "MANUFACTURER_NUMBER"
    NUMBER = "NUMBER"
    ORDER_NAME = "ORDER_NAME"
    ORDER_NUMBER = "ORDER_NUMBER"


class W24IdentifierStakeholder(str, Enum):
    """List of Stakeholders that can be identified by Werk24"""

    SUPPLIER = "SUPPLIER"
    OWNER = "OWNER"
    CUSTOMER = "CUSTOMER"


class W24IdentifierPeriod(str, Enum):
    """List of Period Identifiers that can be identified by Werk24"""

    PREVIOUS = "PREVIOUS"
    CURRENT = "CURRENT"
    FUTURE = "FUTURE"


class W24IdentifierPair(W24CaptionValuePair):
    """Caption-Value Pair for Identifies

    Attributes:
        identifier_type (W24IdentifierType) Type of identifier.
    """

    identifier_type: W24IdentifierType
    stakeholder: Optional[W24IdentifierStakeholder] = None
    period: Optional[W24IdentifierPeriod] = None


class W24FileExtensionType(str, Enum):
    """
    Enum of the extension types.

    For example, pdf and idw extensions will be mapped to
    DRAWING, while step and stl extensions will be mapped
    to MODEL.
    """

    DRAWING = "DRAWING"
    MODEL = "MODEL"
    UNKNOWN = "UNKNOWN"


class W24FilePathType(str, Enum):
    """
    Enum of the file path types, indicating whether a
    POSIX (unix) or WINDOWS path is used. When only a filename
    is indicated, the value will be UNKNOWN
    """

    POSIX = "POSIX"
    WINDOWS = "WINDOWS"
    UNKNOWN = "UNKNOWN"


class W24Filename(W24BaseFeatureModel):
    """
    Object describing all the information that we can
    deduce from a filename that was found on the TitleBlock

    Attributes:

        blurb: Filename and Path as it was found on the drawing.
            Example: /path/to/drawing.pdf

        filename: Filename without the prefix.
            Example drawing.pdf

        extension: Extension of the filename.
            Examples: .pdf, .tar.gz

        extension_type: filetype indicated by the extension. PDF, ...
            extensions are mapped to DRAWING, while STEP, ...
            extensions are mapped to MODEL

        path_type: WINDOWS or POSIX (unix) path types if we have
            found an absolute path. UNKNOWN if we only read
            a filename without prefix.
    """

    blurb: str

    filename: str

    extension: str

    extension_type: W24FileExtensionType

    path_type: W24FilePathType


class W24TitleBlock(BaseModel):
    """
    Information that could be extracted from the Title Block.

    Attributes:

        designation: Designation of the Sheet on the Title Block

        drawing_id: Main Identification Number of the Drawing

        part_ids: List of Part IDs that are located on the drawing.
            Keep in mind that the drawing might define multiple
            variants that each specify their own part id

        reference_ids: List of all the identifiers that we found on the
            drawing. The reference ids will hold all the ids that we found
            on the drawing. These will include both the drawing/part numbers
            as well as all the other ids that we identified in addition,
            such as the Order number, the Item number and so on.

        general_tolerances: General Tolerances quoted on the TitleBlock

        material: Material which is quoted on the TitleBlock

        material_number: Optional Material Number. The material number
            if often present on corporate drawings and typically
            corresponds to a unique material number in the customer's
            material master data. The material number is thus specific
            to the company that maintains the master data and obtaining
            the material specification requires access to that database.

        weight: Weight as read from the TitleBlock. NOTE: this is not
            cross-checked with the material and volume of the part,
            but provided as it was read on the TitleBlock.

        filename_drawing: Filename of the drawing if it is explicitly
            indicated on the title block

        colors: List of colors detected on the TitleBlock or in the
            canvas notes.

        bom_table: Reference to the Bill of Material Table. None
            if you are running on an old client or when no BOM
            table was found.

        general_roughnesses (List[W24GeneralRoughness]): List of the 
            detected general roughnesses.

        reference_roughnesses (List[W24RoughnessReference]): List of the
            detected reference roughnesses.

        unit_specifications (List[W24UnitSpecification]): List of the detected 
            unit specifications. 
    """

    designation: Optional[W24CaptionValuePair]

    drawing_id: Optional[W24CaptionValuePair]
    part_ids: List[W24CaptionValuePair] = []
    reference_ids: List[W24IdentifierPair] = []

    general_tolerances: Optional[W24GeneralTolerances]

    material: Optional[W24Material]
    material_number: Optional[W24CaptionValuePair] = None

    weight: Optional[W24Weight]

    filename_drawing: Optional[W24Filename] = None

    colors: List[W24PropertyColor] = []

    bom_table: Optional[W24BomTable] = None

    general_roughnesses: List[W24GeneralRoughness] = []

    reference_roughnesses: List[W24RoughnessReference] = []

    unit_specifications: List[W24UnitSpecification] = []



""" Command Line Interface for W24 Techread
"""

import traceback
from datetime import datetime
import argparse
from termcolor import colored
from colorama import just_fix_windows_console
import io
import json
import logging
from werk24.exceptions import TechreadException
import os
from collections import namedtuple
from typing import Any, Dict, List, Optional

from dotenv import load_dotenv

from werk24.cli import utils
from werk24.exceptions import RequestTooLargeException
from werk24.models.ask import (
    W24AskCanvasThumbnail,
    W24AskPageThumbnail,
    W24AskPartFamilyCharacterization,
    W24AskSectionalThumbnail,
    W24AskSheetAnonymization,
    W24AskSheetThumbnail,
    W24AskTitleBlock,
    W24AskVariantCAD,
    W24AskVariantExternalDimensions,
    W24AskVariantGDTs,
    W24AskVariantMeasures,
    W24AskVariantRadii,
    W24AskDebug,
    W24AskVariantRoughnesses,
    W24AskVariantThreadElements,
    W24AskVariantProcesses,
)
from werk24.exceptions import ServerException
from werk24.models.techread import (
    W24TechreadException,
    W24TechreadMessageSubtypeError,
    W24TechreadMessageSubtypeProgress,
    W24TechreadMessageType,
)
from werk24.techread_client import LICENSE_LOCATIONS, Hook

# fix the windows console color issue
just_fix_windows_console()

# load the environment variables
for c_location in LICENSE_LOCATIONS:
    if os.path.exists(c_location):
        load_dotenv(c_location)
        break

# set the log level to info for the test setting
# We recommend using logging.WARNING for production
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",  # noqa
    datefmt="%Y-%m-%d %H:%M:%S",
)

# get the local logger
logger = logging.getLogger(__name__)  # pylint:disable = invalid-name


def _open_in_webbrowser(url: str) -> None:
    """Open a URL in the Webbrowser"""
    import webbrowser

    webbrowser.open(str(url), new=2)


# make the configuration of what hooks we want to handle and how
HookConfig = namedtuple("HookConfig", "arg ask function")
hook_config = [
    HookConfig(
        "ask_page_thumbnail",
        W24AskPageThumbnail,
        lambda m: _show_image("Page Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sheet_thumbnail",
        W24AskSheetThumbnail,
        lambda m: _show_image("Sheet Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sheet_anonymization",
        W24AskSheetAnonymization,
        lambda m: _show_image("Sheet Anonymization", m.payload_bytes),
    ),
    HookConfig(
        "ask_canvas_thumbnail",
        W24AskCanvasThumbnail,
        lambda m: _show_image("Canvas Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sectional_thumbnail",
        W24AskSectionalThumbnail,
        lambda m: _show_image("Sectional Thumbnail", m.payload_bytes),
    ),
    # HookConfig(
    #     'ask_variant_angles',
    #     W24AskVariantAngles,
    #     lambda m: _print_payload("Ask Variant Angles", m.payload_dict)),
    HookConfig(
        "ask_variant_external_dimensions",
        W24AskVariantExternalDimensions,
        lambda m: _print_payload("Ask Variant Ext. Dimensions", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_gdts",
        W24AskVariantGDTs,
        lambda m: _print_payload("Ask Variant GDTs", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_measures",
        W24AskVariantMeasures,
        lambda m: _print_payload("Ask Variant Measures", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_radii",
        W24AskVariantRadii,
        lambda m: _print_payload("Ask Variant Radii", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_roughnesses",
        W24AskVariantRoughnesses,
        lambda m: _print_payload("Ask Variant Roughnesses", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_processes",
        W24AskVariantProcesses,
        lambda m: _print_payload("Ask Variant Processes", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_cad",
        W24AskVariantCAD,
        lambda m: _store_variant_cad(m.payload_dict, m.payload_bytes, m.exceptions),
    ),
    HookConfig(
        "ask_titleblock",
        W24AskTitleBlock,
        lambda m: _print_payload("Ask TitleBlock", m.payload_dict),
    ),
    HookConfig(
        "ask_variant_thread_elements",
        W24AskVariantThreadElements,
        lambda m: _print_payload("Ask Variant Thread Elements", m.payload_dict),
    ),
]


def _store_variant_cad(
    payload_dict: Dict[str, Any],
    payload_bytes: Optional[bytes],
    exceptions: List[W24TechreadException],
) -> None:
    """Store the CAD file the current directory

    Args:
    ----
    payload_dict (Dict[str, Any]): Payload Dictionary
    payload_bytes (bytes): CAD that we received as response
    """
    logger.info(f"Ask Variant CAD\n{payload_dict}")

    # print potential exceptions
    if exceptions:
        _log_exceptions(exceptions)

    # store the payload
    if payload_bytes is not None:
        _store_variant_cad_payload(payload_dict, payload_bytes)


def _log_exceptions(exceptions: List[W24TechreadException]) -> None:
    """Log the encountered exceptions as warnings

    Args:
    ----
    exceptions (List[W24TechreadException]): List of
        encountered exceptions. Can be an empty list.
    """
    for cur_exception in exceptions:
        logger.warn(cur_exception)


def _store_variant_cad_payload(
    payload_dict: Dict[str, Any], payload_bytes: bytes
) -> None:
    """Store the CAD file with the knowledge that
    the payload exists

    Args:
    ----
    payload_dict (Dict[str, Any]): Payload Dictionary
    payload_bytes (bytes): CAD response
    """
    # make the filename
    variant_id = payload_dict.get("variant_id")
    filename = f"./w24_ask_variant_cad_{variant_id}.dxf"

    # and write the content
    with open(filename, "wb+") as file_handle:
        file_handle.write(payload_bytes)

    # tell the user
    logger.info(f"CAD response stored in {filename}")


def _get_drawing(file_path: str) -> Optional[io.BufferedReader]:
    """Get the bytes of the file that shall be read.

    Args:
    ----
    file_path (str): path to the input file

    Returns:
    -------
    io.BufferedReader: file handle to the input file
    """
    try:
        return open(file_path, "rb")
    except FileNotFoundError:
        print(f"File {file_path} not found")
        return None


def _print_payload(log_text: str, payload_dict: Dict[str, Any]) -> None:
    """Display the payload dictionary in a format that
    is easy for humans to read

    Args:
    ----
    log_text (str): Headline info
    payload_dict (Dict[str, Any]): Payload dictionary
    """
    print(log_text)
    payload_json = json.dumps(payload_dict, indent=4)
    print(payload_json)


def _show_image(log_text: str, image_bytes: bytes) -> None:
    """Display a debug image. The function relies on
    the PIL functionality of Image.show(). The
    actual behavior will thus be different across platforms

    Args:
    ----
    log_text (str): Log Text that we want to write back
        to the command line interface
    image_bytes (bytes): byte representation of the image
        that is to be displayed
    """

    # Import the PIL at runtime to make sure that
    # you can still use the client without the need
    # to install Pillow
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
    except ImportError:
        logger.warning(
            "Viewing image-like responses requires the installation"
            " of the PIL package: pip install pillow. Image skipped."
        )
        return

    # otherwise print the text
    logger.info(log_text)

    # and show the image
    image = Image.open(io.BytesIO(image_bytes))
    try:
        image.show(title=log_text)
    except BaseException:
        logger.warning(
            "Image cannot be displayed. " "Please check your local security settings."
        )
        return


async def main(args: argparse.Namespace) -> None:
    """
    Run the Techread command with the CLI arguments

    Args:
    ----
    args(argparse.Namespace): CLI args generated by
        argparse
    """
    try:
        # make the hooks from the arguments
        hooks = _make_hooks_from_args(args)

        # get the client instance and handle
        # potential errors
        client = utils.make_client()

        # get the drawing
        drawing = _get_drawing(args.input_file)
        if drawing is None:
            return

        async with client as session:
            await session.read_drawing_with_hooks(drawing, hooks, sub_account=None)
            drawing.close()

    except TechreadException as exception:
        print()
        print(colored("=" * 3 + "[Techread Error]" + "=" * 61, "red"))

        print(colored(exception.cli_message_header, "red"))
        print("-" * 80)

        print(colored(exception.cli_message_body, "yellow"))
        print(colored("=" * 80, "red"))

    except BaseException as exception:
        catch_crash(exception)


def catch_crash(exception: BaseException) -> None:
    """
    Catch crashes during program execution and log them in a file.

    Args:
    ----
    exception (BaseException): The exception instance captured during crash.

    Returns:
    -------
    None
    """
    filename = write_crash_log()
    print_crash_message(filename)


def write_crash_log() -> str:
    """
    Create a log entry of an internal error that occurred during program execution.

    Returns:
    -------
    filename (str): The name of the log file where the error details are written.
    """

    timestamp = str(datetime.now().isoformat()).replace(":", "-")
    filename = f"werk24-crash-{timestamp}.log"
    content = f"=== [Internal Error] ===\n\nCALLSTACK:\n{traceback.format_exc()}"

    with open(filename, "w") as crash_log:
        crash_log.write(content)

    return filename


def print_crash_message(filename: str) -> None:
    """
    Print an error message regarding an internal crash.

    Args:
    ----
    filename (str): The name of the log file containing the error details.

    Returns:
    -------
    None
    """
    separator_line = colored("=" * 80, "red")

    crash_message = [
        separator_line,
        "",
        "=== [Internal Error] ===",
        "An internal error occurred.",
        "To report the error, please send the file called:",
        "",
        f"  --> {filename} <-- ",
        "",
        "to info@werk24.io",
        "We apologize for the inconvenience.",
        "",
        separator_line,
    ]
    print("\n".join(colored(li, "red") for li in crash_message))


def _make_hooks_from_args(args: argparse.Namespace) -> List[Hook]:
    # add a general hook to deal with internal errors
    hook_error_internal = Hook(
        message_type=W24TechreadMessageType.ERROR,
        message_subtype=W24TechreadMessageSubtypeError.INTERNAL,
        function=lambda msg: logger.error("Internal Error %s", msg.payload_dict),
    )

    # add a general hook to deal with PROGRESS messages
    hook_progress_started = Hook(
        message_type=W24TechreadMessageType.PROGRESS,
        message_subtype=W24TechreadMessageSubtypeProgress.STARTED,
        function=lambda msg: logger.info("Techread process started"),
    )

    # tell the api what asks you are interested in,
    # and define what to do when you receive the result
    hooks = [
        Hook(ask=c.ask(), function=c.function)
        for c in hook_config
        if getattr(args, c.arg)
    ] + [hook_error_internal, hook_progress_started]

    # take special care of teh W24AskPartFamilyCharacterization
    if args.part_family_id is not None:
        c_hook = Hook(
            ask=W24AskPartFamilyCharacterization(part_family_id=args.part_family_id),
            function=lambda m: _print_payload(
                "Part Family Characterization", m.payload_dict
            ),
        )
        hooks.append(c_hook)

    if args.debug_key is not None:
        c_hook = Hook(
            ask=W24AskDebug(debug_key=args.debug_key),
            function=lambda m: _open_in_webbrowser(m.payload_url),
        )
        hooks.append(c_hook)

    return hooks



from typing import Literal

from werk24.models.property.base import W24Property
from werk24.models.value import W24PhysicalQuantity


class W24PropertyWeight(W24Property):
    """Weight of the Part.

    NOTE: this might in the future be extended by another
    attribute that indicates whether the weight was measured
    or calculated.

    Attributes:
        value: weight in the units indicated on the drawing.
    """
    property_type: Literal["WEIGHT"] = "WEIGHT"
    property_subtype: Literal["WEIGHT"] = "WEIGHT"
    weight: W24PhysicalQuantity



import argparse
import platform
import sys
from datetime import datetime
import os.path
import websockets
from websockets.exceptions import InvalidStatusCode
from werk24.cli import utils
import werk24
from werk24.techread_client import LicenseError
from werk24.exceptions import UnauthorizedException

LINE_LENGTH = 80
URI = "wss://ws-api.w24.co/v2"


async def main(args: argparse.Namespace) -> None:
    """
    Perform the Health Checks.
    Args:
        args (argparse.Namespace): Command-line arguments.
    """
    print_section("System Information", get_system_info())
    print_section("License Information", get_license_info())
    print_section("Module Information", get_module_info())
    print_section("Network Information", await get_network_info())
    print_section("Miscellaneous", get_misc_info())


def print_section(headline: str, information: dict) -> None:
    """
    Print a section with given headline and information.
    Args:
        headline (str): The headline of the section.
        information (dict): A dictionary containing the information to be printed.
    """
    filler = max(0, LINE_LENGTH - 9 - len(headline))
    print("\n{0} [{1}] {2}".format("-" * 5, headline, "-" * filler))
    for caption, value in information.items():
        print("{0:<30}: {1}".format(caption, value))
    print("-" * LINE_LENGTH)


def get_system_info() -> dict:
    """Get system information.
    Returns:
        dict: A dictionary containing the system information.
    """
    return {
        "OS": f"{platform.system()} {platform.release()}",
        "Python Version": sys.version,
    }


def get_license_info() -> dict:
    """Get the license Information

    Returns:
    -------
    dict: A dictionary containing the license information.
    """
    try:
        utils.make_client()

    except LicenseError:
        return {"License Information": "NOT FOUND"}

    except UnauthorizedException:
        return {"License Information": "UNAUTHORIZED"}

    return {
        "License Information": "FOUND",
    }


def get_module_info() -> dict:
    """Get module information.
    Returns:
        dict: A dictionary containing the module information.
    """
    return {
        "Werk24 Version": werk24.__version__,
    }


async def get_network_info() -> dict:
    """Test connection and get network info.
    Returns:
        dict: A dictionary containing the results of the connection test.
    """
    try:
        async with websockets.connect(URI):
            return {"ws-api.w24.co Connection": "Successful"}
    except InvalidStatusCode as e:
        if e.status_code == 401:
            return {"Websocket Connection": "Successful, Connected"}
        else:
            return {"Websocket Connection": f"Unsuccessful: {str(e.status_code)}"}
    except Exception as e:
        return {"Websocket Connection": f"Unsuccessful: {type(e)}, {e}"}


def get_misc_info() -> dict:
    def _check_werk24_name_conflict() -> str:
        if os.path.exists("werk24.py"):
            return "! NAME CONFLICT ! Please rename your werk24.py file"
        else:
            return "No Conflict"

    return {
        "Time": datetime.now(),
        "Werk24 Name Conflict": _check_werk24_name_conflict(),
    }



from werk24.models.feature.base import W24Feature
from typing import Literal
from werk24.models.size import W24Size
from werk24.models.tolerance import W24ToleranceType
from werk24.models.angle import W24AngleLabel


class W24FeatureChamfer(W24Feature):
    type: Literal["CHAMFER"] = "CHAMFER"
    size: W24Size
    size_tolerance: W24ToleranceType
    angle: W24AngleLabel
    angle_tolerance: W24ToleranceType



from decimal import Decimal
from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, validator

from werk24.models.unit import W24UnitLength

from .base_feature import W24BaseFeatureModel


class W24GeneralTolerancesStandard(str, Enum):
    """Enum of all supported
    General Tolerance Standards
    """

    DIN_7168 = "DIN 7168"
    ISO_2768 = "ISO 2768"
    ISO_4759_1 = "ISO 4759-1"
    TOLERANCE_NOTE = "TOLERANCE_NOTE"


class W24ToleranceProperty(str, Enum):
    """Enum of all attributes that can
    be described by general tolerances
    """

    ANGULAR = "ANGULAR"
    FLATNESS = "FLATNESS"
    LINEAR = "LINEAR"
    PERPENDICULARITY = "PERPENDICULARITY"
    RADIUS = "RADIUS"
    RUNOUT = "RUNOUT"
    STRAIGHTNESS = "STRAIGHTNESS"
    SYMMETRY = "SYMMETRY"


class W24GeneralTolerancesPrinciple(str, Enum):
    """Enum of the supported General Tolerance
    Principles.
    """

    INDEPENDENCE = "INDEPENDENCE"
    ENVELOPE = "ENVELOPE"


class W24IntervalEnd(str, Enum):
    """
    Enum for handling the interval ends.
    Open for  `<`
    Close for `<=`
    """

    OPEN = "OPEN"
    CLOSE = "CLOSE"


class W24ToleranceTableItem(BaseModel):
    """
    Tolerance Table Item.

    Attributes:
        nominal_min: Lower bound of the nominal value
        nominal_min_end: Interval end of the lower bound
        nominal_max: Upper bound of the nominal value
        nominal_max_end: Interval end of the upper bound
        decimal_places: Number of decimal places. This is
            required for US-type tolerances.
        deviation_min: Lower bound of the deviation
        deviation_max: Upper bound of the deviation
    """

    nominal_min: Optional[Decimal]
    nominal_min_end: W24IntervalEnd = W24IntervalEnd.OPEN
    nominal_max: Optional[Decimal]
    nominal_max_end: W24IntervalEnd = W24IntervalEnd.CLOSE

    decimal_places: Optional[int] = None

    deviation_min: Optional[Decimal]
    deviation_max: Optional[Decimal]

    unit: Optional[W24UnitLength] = None

    @validator("nominal_min", pre=True)
    def nominal_min_validator(  # NOQA
        cls, raw: Union[str, float, None]
    ) -> Optional[Decimal]:
        """Ensure the proper conversion of the the
        Decimal value
        """
        return cls._convert_decimal(raw)

    @validator("nominal_max", pre=True)
    def nominal_max_validator(  # NOQA
        cls, raw: Union[str, float, None]
    ) -> Optional[Decimal]:
        """Ensure the proper conversion of the the
        Decimal value
        """
        return cls._convert_decimal(raw)

    @staticmethod
    def _convert_decimal(raw: Union[str, float, None]) -> Optional[Decimal]:
        """Handle the decimal conversion

        Args:
            raw (Union[str, float, None]): Raw value

        Raises:
            ValueError: raised when the data type
                does not match the understood types

        Returns:
            Optional[Decimal]: None if the raw value
                corresponds to -Infinity, +Infinity or
                NaN. Decimal('0') if Decimal('-0') or
                the true value
        """

        # ensure we are working with
        # the correct type
        if isinstance(raw, str) or isinstance(raw, float) or isinstance(raw, int):
            decimal = Decimal(raw)

        # accept decimal
        elif isinstance(raw, Decimal):
            decimal = raw

        # allow None values
        elif raw is None:
            return None

        # reject the rest
        else:
            data_type = type(raw)
            raise ValueError(f"Unsupported datatype '{data_type}'")

        # interpret the values
        # handle the special values
        if decimal in {Decimal("Infinity"), Decimal("-Infinity"), Decimal("NaN")}:
            return None

        # enforce zero to be positive
        if decimal == Decimal("-0"):
            return Decimal("0")

        # accept the value
        return decimal


class W24ToleranceClass(BaseModel):
    """Tolerance Class which matches an individual attribute
    of the General Tolerances to a tolerance property and
    tolerance table

    Attributes:
        blurb: Tolerance class label for human consumption

        raw_ocr_blurb: Tolerance note text as annotated on
            the drawing.

        property: Property that is being tolerated

        table: Rows of the tolerance table that correspond
            to the selected tolerance class
    """

    blurb: str
    raw_ocr_blurb: str = ""
    property: W24ToleranceProperty
    table: List[W24ToleranceTableItem]


class W24GeneralTolerances(W24BaseFeatureModel):
    """
    General Tolerances on the Title Block of the Technical Drawing.

    Attributes:
        blurb: Blurb of the general tolerances for human consumption.
        tolerance_standard: General Tolerance Standard that was defined
            in the Drawing
        principle: Principle that is annotated on the general tolerance
            by "-E" (or the lack of if).
        angular_class: Angular toleration class
        flatness_class: Flatness toleration class
        straightness_class: Straightness toleration class
        linear_class: Linear toleration class
        radius_class: Radius and chamfer toleration class
        symmetry_class: Symmetry toleration class
        perpendicularity_class: Perpendicularity toleration class

    """

    blurb: str
    tolerance_standard: W24GeneralTolerancesStandard
    principle: Optional[W24GeneralTolerancesPrinciple] = None
    angular_class: Optional[W24ToleranceClass] = None
    flatness_class: Optional[W24ToleranceClass] = None
    straightness_class: Optional[W24ToleranceClass] = None
    linear_class: Optional[W24ToleranceClass] = None
    radius_class: Optional[W24ToleranceClass] = None
    runout_class: Optional[W24ToleranceClass] = None
    symmetry_class: Optional[W24ToleranceClass] = None
    perpendicularity_class: Optional[W24ToleranceClass] = None



""" Defintion of all the W24Radius class its support structures
"""

from typing import Optional, Union
from enum import Enum

from pydantic import UUID4, BaseModel, validator

from .base_feature import W24BaseFeatureModel
from .size import W24Size
from .tolerance import (
    W24Tolerance,
    W24ToleranceApproximation,
    W24ToleranceFitsizeISO,
    W24ToleranceGeneral,
    W24ToleranceMaximum,
    W24ToleranceMinimum,
    W24ToleranceOffSize,
    W24ToleranceReference,
    W24ToleranceTheoreticallyExact,
)
from .unit import W24UnitLength


class W24CurvatureType(str, Enum):
    """Curvature types of Radius"""

    CONCAVE = "CONCAVE"
    CONVEX = "CONVEX"
    PLANE = "PLANE"


class W24RadiusLabel(BaseModel):
    """Radius Label

    Attributes:
    ----------
        blurb: String representation of the Radius for human consumption

        curvature_type: type of radius curvature. Mostly used in optical
            radii. It can be concave or convex. If no information is
            available, default to None.

        quantity: Quantity of the annotated radius, e.g., 2 x R4 returns
            quantity=2

        quality: Deprecated. This was a typo of the quantity.

        size: Size of the Radius as referred in the drawing.

        size_tolerance: Tolerance details of the Radius. Please keep in
            mind that Radii can carry special tolerances. If none are
            mentioned on the drawing, the general tolerances apply.

        unit: Length units of the size and size_tolerance. In most cases this
            will be be millimeter (METRIC) or inch (IMPERIAL) and be consistent
            for the complete drawing. Exceptions are very rare, but exist.
    """

    @validator("size_tolerance", pre=True)
    def deserialize_size_tolerance(cls, v):
        if isinstance(v, W24Tolerance):
            return v
        return W24Tolerance.parse_obj(v)

    blurb: str

    curvature_type: Optional[W24CurvatureType] = None

    quantity: int = 1
    quality: int = 1

    size: W24Size

    size_tolerance: Union[
        W24ToleranceFitsizeISO,
        W24ToleranceReference,
        W24ToleranceOffSize,
        W24ToleranceGeneral,
        W24ToleranceTheoreticallyExact,
        W24ToleranceMinimum,
        W24ToleranceMaximum,
        W24ToleranceApproximation,
        W24ToleranceGeneral,
    ] = W24ToleranceGeneral()

    unit: Optional[W24UnitLength] = None


class W24Radius(W24BaseFeatureModel):
    """Radius Feature

    Attributes:
        radius_id: Unique UUID4 identifier. This can be used to provide
            automated feedback about customer changes.

        label: Label of the radius.

        confidence: Werk24 calcualtes an internal confidence score for
            reach radius. Depending on your use-case, you might want
            to consider or discard low-confidence radii. This value
            allows you to do so. The value ranges from 0.0 to 1.0
    """

    radius_id: UUID4

    label: W24RadiusLabel

    confidence: float



from typing import Literal

from werk24.models.property.base import W24Property


class W24PropertyColor(W24Property):
    """Color Property.

    NOTE: at this stage we do not make
        the colors comparable. There
        might be ways of doing this, but
        colors are complex and most
        color systems cannot express the
        complete space (e.g., RGB is not
        describing the glossiness of the
        color).
    """
    property_type: Literal["COLOR"] = "COLOR"


class W24PropertyColorRAL(W24PropertyColor):
    """Color Property as RAL number.

    Attributes:
        ral_number (str): RAL number of the
            color. Note that there are three
            different RAL numbering systems
            in use.
    """
    property_subtype: Literal["RAL"] = "RAL"
    ral_number: str


class W24PropertyColorBlurb(W24PropertyColor):
    """Color Property as Blurb.

    Attributes:
        color_name (str): English Color name.
            See Documentation for full list of
            supported colors.
    """
    property_subtype: Literal["BLURB"] = "BLURB"
    color_name: str



"""Improvement of the Pydantic Deserializer.

This allows us to dynamically specify the
models without the need to update the deserializer.

Be sure to specify the discriminators in the Config.

Example:
-------

class MetaData(W24TypedModel):
    class Config:
        discriminators = ('meta_data_type',)

class MetaDataLogo(MetaData):
    meta_data_type: Literal["LOGO"] = "LOGO"

class MetaDataDesignation(MetaData):
    meta_data_type: Literal["DESIGNATION"] = "DESIGNATION"
    designation: str


# serializiation
obj = MetaDataDesignation(blurb="a", designation="b").dict()

# deserialization
print(MetaData.parse_obj(obj))

"""
from pydantic import BaseModel
from pint import Quantity
from pydantic_core import PydanticUndefined

class W24TypedModel(BaseModel):
    """
    Parent to any type-based sub-model.

    Attributes:
    ----------
    _subtypes_ (Dict[Tuple[str,...], BaseModel]): Map
        that associates the discriminators to a specific
        model that shall be called.
    """

    __subtypes__ = {}

    class Config:
        arbitrary_types_allowed = True

        """Have the custom encoders here.
        This is not the nicest solution, but more
        a workaround until Pydantic 2.0 is ready.
        See: https://github.com/pydantic/pydantic/issues/2277
        """
        json_encoders = {
            # NOTE: specify a custom validator to make
            # sure that the serialization is done correctly.
            # See validator for details.
            Quantity: lambda v: str(v)
        }

    @classmethod
    def __pydantic_init_subclass__(
        cls,
    ):
        """Called when a subclass is specified.

        Registers the class locally.
        """
        key_ = tuple(
            [cls._first_child()]
            + [
                None
                if cls.model_fields[disc].default == PydanticUndefined
                else cls.model_fields[disc].default
                for disc in cls.Config.discriminators
            ]
        )
        cls.__subtypes__[key_] = cls

    @classmethod
    def _first_child(cls):
        parent = cls
        for _ in range(100):
            child = parent
            parent = parent.__bases__[0]
            if parent.__name__ == "W24TypedModel":
                return child

        raise RecursionError()

    @classmethod
    def _convert_to_real_type_(cls, data):
        """Convert the data to the correct subtype."""
        # get the key from the data.
        key_ = tuple([cls] + [data.get(disc) for disc in cls.Config.discriminators])

        # check whether the subtype actually exists.
        # Be careful with updates here.
        sub = cls.__subtypes__.get(key_)
        if sub is None:
            raise TypeError(f"Unsupported sub-type: {key_}")

        # parse the object using the subclass
        return sub(**data)

    @classmethod
    def parse_obj(cls, obj):
        """Parse the object with the correct deserializer."""
        return cls._convert_to_real_type_(obj)



""" Command Line Interface for W24
"""
import argparse
import asyncio

import werk24.cli.health_check
import werk24.cli.support
import werk24.cli.techread


def main() -> None:
    parser = argparse.ArgumentParser(prog="w24cli")
    subparsers = parser.add_subparsers(
        dest="service", help="Service. Currently supported: techread, auth, support"
    )
    subparsers.required = True

    _add_support_parser(subparsers)
    _add_techread_parser(subparsers)

    args = parser.parse_args()
    if args.service == "techread":
        asyncio.run(werk24.cli.techread.main(args))

    elif args.service == "support":
        asyncio.run(werk24.cli.support.main(args))

    elif args.service == "health_check":
        asyncio.run(werk24.cli.health_check.main(args))


def _add_support_parser(subparsers):
    support_parser = subparsers.add_parser("support", help="Support Service")

    support_subparsers = support_parser.add_subparsers(
        dest="command",
        help="Command. Currently supported: create-helpdesk-task",
        required=True,
    )

    create_parser = support_subparsers.add_parser(
        "create-helpdesk-task", help="Create a new help desk task"
    )

    create_parser.add_argument("--request-id", action="store", required=True)
    create_parser.add_argument("--observed-outcome", action="store", required=True)
    create_parser.add_argument("--expected-outcome", action="store", required=True)
    create_parser.add_argument("--comment", action="store")
    create_parser.add_argument("--importance", action="store", required=True)


def _add_techread_parser(subparsers):
    parser_techread = subparsers.add_parser(
        "techread", help="Submit a Technical Drawing to Werk24 for analysis"
    )

    parser_techread.add_argument(
        "input_file", help="path to the file that is to be analyzed"
    )

    parser_techread.add_argument(
        "--sub-account",
        help="Sub-Account that the request should be attributed to.",
        action="store",
        dest="sub_account",
        default=None,
    )

    parser_techread.add_argument(
        "--ask-techread-started",
        help="ask for a callback when the techread process has been picked up by a worker",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-page-thumbnail",
        help="ask for a thumbnail for each page in the document",
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-sheet-thumbnail",
        help="ask for a thumbnail of each sheet in the document",
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-sheet-anonymization",
        help="ask for an anonymized of each sheet in the document",
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-canvas-thumbnail",
        help="ask for a thumbnail of each canvas in the document",
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-sectional-thumbnail",
        help="ask for a thumbnail of each sectional of each sheet in the document",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-cad",
        help="ask for the CAD file of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-external-dimensions",
        help="ask for the External Dimensions of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-gdts",
        help="ask for the GD&Ts of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-measures",
        help="ask for the measures of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-processes",
        help="ask for the Proceses of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-roughnesses",
        help="ask for the Surface Roughnesses of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-radii",
        help="ask for the Radii of each variant",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-variant-thread-elements",
        help="ask for the Thread Elements",  # noqa
        action="store_true",
    )

    parser_techread.add_argument(
        "--ask-titleblock", help="ask for the Title Block", action="store_true"  # noqa
    )

    parser_techread.add_argument(
        "--ask-part-family-characterization",
        help="ask for the Part Family Characterization",  # noqa
        action="store",
        dest="part_family_id",
    )

    parser_techread.add_argument(
        "--ask-debug",
        help="internal debug option",
        action="store",
        dest="debug_key",
    )

    parser_health_check = subparsers.add_parser(
        "health_check", help="Perform Health Checks"
    )
    parser_health_check.add_argument("type", help="Type of Health Check to perform")


if __name__ == "__main__":
    main()



from decimal import Decimal
from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel

from werk24.models.standard import W24Standard
from werk24.models.typed_model import W24TypedModel

from .base_feature import W24BaseFeatureModel
from .gender import W24Gender


class W24Tolerance(W24TypedModel):
    """Abstract Base Class to cover the Tolerances.

    Attributes:
        blurb (str): String representation for human consumption

        toleration_type (W24SizeToleranceType):  Toleration Type for
            deserialization
    """

    class Config:
        discriminators = ("toleration_type",)

    blurb: str
    toleration_type: str


class W24ToleranceGradeWarning(str, Enum):
    """Warnings associated with the Tolerance Grade."""

    SIZE_LARGER_THAN_NORM = "SIZE_LARGER_THAN_NORM"
    TOLERANCE_WIDTH_SMALLER_THAN_NORM = "TOLERANCE_WIDTH_SMALLER_THAN_NORM"
    TOLERANCE_WIDTH_LARGER_THAN_NORM = "TOLERANCE_WIDTH_LARGER_THAN_NORM"


class W24ToleranceGrade(BaseModel):
    """Tolerance Grade following ISO 286-1

    Attributes:
        blurb (Optional[str]): Blurb for human consumption.

        grade (Optional[str]): Tolerance Grade in the range
            of IT01 to IT18. None if the tolerance is outside the
            normed region

        warning (Optional[W24ToleranceGradeWarning]): the norm is
            limited to the size range (0 - 3150mm) and the
            tolerance range of IT1 - IT18. When we reach the
            ends of this norm, we return a warning.

        NOTE: when a tolerance is outside the norm, we still
            return the closest matching grade.

        NOTE: when tolerances are implausible (e.g., 3+/-6),
            Werk24 will strip the tolerance completely and
            return an untoleranced measure of nominal size 3.

    """

    blurb: str

    grade: Optional[str]

    warning: Optional[W24ToleranceGradeWarning]


class W24ToleranceFitsizeISO(W24Tolerance):
    """ISO fit size tolerances

    Attributes:
        blurb (str): Text representation for human consumption.

        deviation_lower (Decimal): Lower deviation from the
            nominal size

        deviation_upper (Decimal): Upper deviation from the
            nominal size

        fundamental_deviation (str): Fundamental deviation of
            the fit (e.g., 'H' for a 'H7' fit)

        tolerance_grade (Optional[int]): Tolerance Grade corresponding
            to ISO 286-1. In German IT-Grad.
    """

    toleration_type: str = "FIT_SIZE_ISO"

    deviation_lower: Decimal

    deviation_upper: Decimal

    fundamental_deviation: str

    tolerance_grade: W24ToleranceGrade

    @property
    def deviation_width(self) -> Decimal:
        """Deviation Width

        Returns:
            Decimal: Deviation width
        """
        return self.deviation_upper - self.deviation_lower


class W24ToleranceReference(W24Tolerance):
    """Measures written in brackets are Reference."""

    toleration_type: str = "REFERENCE"


class W24ToleranceOffSize(W24Tolerance):
    """Off-size based tolerances

    Attributes:
        blurb (str): Text representation for human consumption.

        deviation_lower (Decimal): Lower deviation from the
            nominal size

        deviation_upper (Decimal): Upper deviation from the
            nominal size

        tolerance_grade (int): Tolerance Grade corresponding to
            ISO 286-1. In German IT-Grad.
    """

    toleration_type: str = "OFF_SIZE"
    deviation_lower: Decimal
    deviation_upper: Decimal
    tolerance_grade: W24ToleranceGrade


class W24ToleranceGeneral(W24Tolerance):
    """General Tolerances

    Attributes:
        tolerance_type: W24SizeToleranceType General Tolerances
            as key to differentiate between the different
            tolerance types

        blurb: Blurb of the General Tolerance following the
            pattern of W24ToleranceOffSize

        standard: GeneralToleranceStandard that is specified in
            the TitleBlock. If the TitleBlock is not requested
            of it no GeneralTolerance was found, this is set to
            None.

        standard_class: GeneralTolerance Class if known. None
            otherwise.

        deviation_lower: Lower deviation from the nominal size
            if the General Tolerance is known. None otherwise.

        deviation_upper: Upper deviation from the nominal size
            if the General Tolerance is known. None otherwise.

        tolerance_grade (int): Tolerance Grade corresponding to
            ISO 286-1. (German: IT-Grad).
    """

    toleration_type: str = "GENERAL_TOLERANCES"
    blurb: str = ""
    standard: Optional[W24Standard] = None
    standard_class: Optional[str] = None
    deviation_lower: Optional[Decimal] = None
    deviation_upper: Optional[Decimal] = None
    tolerance_grade: Optional[W24ToleranceGrade] = None


class W24ToleranceTheoreticallyExact(W24Tolerance):
    """Theoretically Exact Measures after ISO 5458
    must not be tolerated. They are indicated by a small
    rectangular frame.

    Example:
        +------+
        |  15  |
        +------+

    !!! note
        In practice, we see Technical drawings contain
        tolerated, theoretically exact measures.
        Example:
            +------------+
            | 15 +/- 0.2 |
            +------------+
        In these situations the toleration takes priority.
    """

    toleration_type: str = "THEORETICALLY_EXACT"


class W24ToleranceMinimum(W24Tolerance):
    """Minimum Size of a measure
    Example:
        min. 15
    """

    toleration_type: str = "MINIMUM"


class W24ToleranceMaximum(W24Tolerance):
    """Maximum Size of a measure
    Example:
        max 15
    """

    toleration_type: str = "MAXIMUM"


class W24ToleranceApproximation(W24Tolerance):
    """Approximation of a measure
    Example:
        ~ 15
        ca. 14
    """

    toleration_type: str = "APPROXIMATION"


class W24ToleranceFeature(W24BaseFeatureModel):
    """Characterization of a Tolerance Feature.

    Attributes:
        gender: Gender (male or female) of the tolerance feature.
            This is determined by checking whether the tolerance feature is
            located on the outer contour of the part or inside the part.
            When the outer contour is unavailable (e.g., in detail drawings),
            the gender is set to None.

        length: Length of the slug corresponding to the tolerance feature.

    """

    gender: Optional[W24Gender]

    length: Optional[Decimal]

    tolerance: W24Tolerance


W24ToleranceType = Union[
    W24ToleranceApproximation,
    W24ToleranceFitsizeISO,
    W24ToleranceGeneral,
    W24ToleranceMaximum,
    W24ToleranceMinimum,
    W24ToleranceOffSize,
    W24ToleranceReference,
    W24ToleranceTheoreticallyExact,
]
